#ifndef CTESK_UTIL_HPP
#define CTESK_UTIL_HPP

#include <stdexcept>
#include <utility>
#include <iterator>


namespace cpptesk {
	namespace utils {

template<typename T>
inline T const* _ignore(const T& x) { return &x;}

		namespace _iterator_helper {
template<typename C>
struct iterator_category_order;

template<>
struct iterator_category_order<std::input_iterator_tag>
{
	typedef struct { static const int value = 0; } type;
	static const int value = type::value;
};

template<>
struct iterator_category_order<std::forward_iterator_tag>
{
	typedef struct { static const int value = 1; } type;
	static const int value = type::value;
};

template<>
struct iterator_category_order<std::bidirectional_iterator_tag>
{
	typedef struct { static const int value = 2; } type;
	static const int value = type::value;
};

template<>
struct iterator_category_order<std::random_access_iterator_tag>
{
	typedef struct { static const int value = 3; } type;
	static const int value = type::value;
};

template<int diff, typename C1, typename C2>
struct minimal_category_impl;

template<typename C1, typename C2> struct minimal_category_impl<3,C1,C2> { typedef C2 type; };
template<typename C1, typename C2> struct minimal_category_impl<2,C1,C2> { typedef C2 type; };
template<typename C1, typename C2> struct minimal_category_impl<1,C1,C2> { typedef C2 type; };
template<typename C1, typename C2> struct minimal_category_impl<0,C1,C2> { typedef C2 type; };

template<typename C1, typename C2> struct minimal_category_impl<-1,C1,C2> { typedef C1 type; };
template<typename C1, typename C2> struct minimal_category_impl<-2,C1,C2> { typedef C1 type; };
template<typename C1, typename C2> struct minimal_category_impl<-3,C1,C2> { typedef C1 type; };

template <typename C1, typename C2>
struct minimal_iterator_category
{
	typedef typename minimal_category_impl<
		iterator_category_order<C1>::value - iterator_category_order<C2>::value,
		C1,
		C2
	>::type type;
};
		} // ctesk::utils::_iterator_helper

template <typename I1, typename I2>
struct InputIteratorPair {
	I1 i1;
	I2 i2;
//	typedef std::forward_iterator_tag iterator_category;
	//typedef typename _iterator_helper::minimal_iterator_category<
	//	typename std::iterator_traits<I1>::iterator_category,
	//	typename std::iterator_traits<I2>::iterator_category
	//>::type iterator_category;

	typedef std::input_iterator_tag iterator_category;
	typedef typename std::iterator_traits<I1>::value_type T1;
	typedef typename std::iterator_traits<I2>::value_type T2;
	typedef std::pair<T1, T2> value_type;
	typedef size_t difference_type;
	typedef difference_type distance_type;	// retained
	typedef value_type* pointer;
	typedef value_type& reference;

	InputIteratorPair() : i1(), i2(0) {}
	InputIteratorPair(const I1& i1, const I2& i2) : i1(i1), i2(i2) {}
	// TODO    -     ,    
	//  
	value_type operator* () { std::pair<T1, T2> result; result.first = *i1; result.second = *i2; return result; }

	InputIteratorPair& operator++ () { ++i1; ++i2; return *this; }
	InputIteratorPair operator++ (int) { InputIteratorPair result(*this); ++i1; ++i2; return result; }

	template<typename IU, typename IV>
	bool operator == (const InputIteratorPair<IU, IV>& other) const { return i1 == other.i1 && i2 == other.i2; }
	template<typename IU, typename IV>
	bool operator != (const InputIteratorPair<IU, IV>& other) const { return !(*this == other); }
};

template <class I>
struct Range
{
  typedef I Iterator;
private:
  I _begin;
  I _end;
  size_t _size;
public:
  Range() : _begin(I()), _end(I()), _size() {}
  Range(const I& x, const I& y, size_t size) : _begin(x), _end(y), _size(size) {}
  template <class U>
    Range (const Range<U> &r) : _begin(r._begin), _end(r._end), _size(r._size) { }
  I begin() { return _begin; }
  I end() { return _end; }
  size_t size() { return _size(); }

};

/// @brief Auxiliary class used to identify the current class type.
/// @detailed This class introduces <tt>this_class_t</tt> identifier that is a
/// <tt>typedef</tt> to the type parameter <tt>T</tt>.
/// @tparam T any class.
template <typename T>
struct ThisType
{
    typedef T this_class_t;
};

/// @brief Auxiliary class used to identify the current class type and the type of its parent.
/// @detailed This class introduces <tt>this_class_t</tt> and <tt>parent_class_t</tt> member types.
/// <tt>this_slass_t</tt> is the type of the class, and <tt>parent_class_t</tt> is the type of the parent class.
/// @tparam ThisType any class.
/// @tparam ParentType the parent class of <tt>ThisType</tt>.
template <typename ThisType, typename ParentType>
struct DerivedType
{
    typedef ThisType this_class_t;
    typedef ParentType parent_class_t;
};

/// @brief Registers a routine that will be called upon exit of the embracing block.
/// This class followed the pattern of automatic resource management in C++.
///
/// @tparam F type of callable objects - function of functor.
template <typename F>
class ExecuteOnBlockEnd
{
    public:
    ExecuteOnBlockEnd(F f) : _f(f), _cancelled(false) {}
    ~ExecuteOnBlockEnd() { if (!_cancelled) _f(); }

    void cancel() { _cancelled = true; }
    bool isCancelled() { return _cancelled; }

    private:
    ExecuteOnBlockEnd(const ExecuteOnBlockEnd&);
    ExecuteOnBlockEnd& operator = (ExecuteOnBlockEnd&);

    F _f;
    bool _cancelled;
};

	}
}

#endif // CTESK_UTIL_HPP
