/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Debug printing facilities 
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>

#include <stdio.h>

#include <iostream>
#include <sstream>
#include <string>

///////////////////////////////////////////////////////////////////////////////
// Debug
///////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_DEBUG_BUFFER_SIZE 1024

#define CPPTESK_DEBUG_PRINT(debug_level, debug_message) \
    do { \
        std::stringstream __cpptesk_buffer; \
        __cpptesk_buffer << debug_message; \
        ::cpptesk::hw::Debug::print(debug_level, __cpptesk_buffer.str()); \
    } while(false)

#define CPPTESK_DEBUG_PRINTF(debug_level, debug_message, ...) \
    do { \
        char __cpptesk_buffer[__CPPTESK_DEBUG_BUFFER_SIZE]; \
        if(snprintf(__cpptesk_buffer, __CPPTESK_DEBUG_BUFFER_SIZE, debug_message, __VA_ARGS__) >= __CPPTESK_DEBUG_BUFFER_SIZE) { \
            CPPTESK_ASSERTION(false, "Too long debug message"); \
        } else { \
            ::cpptesk::hw::Debug::print(debug_level, __cpptesk_buffer); \
        } \
    } while(false)

#define __CPPTESK_DEBUG_PRINT_FILE_LINE(debug_level, debug_message, file, line) \
    CPPTESK_DEBUG_PRINT(debug_level, debug_message << " (" << file << ", " << std::dec << line << ")")

#define __CPPTESK_DEBUG_PRINTF_FILE_LINE(debug_level, debug_message, file, line, ...) \
    CPPTESK_DEBUG_PRINTF(debug_level, debug_message " (%s, %d)", __VA_ARGS__, file, line)

#define CPPTESK_DEBUG_PRINT_FILE_LINE(debug_level, debug_message) \
    __CPPTESK_DEBUG_PRINT_FILE_LINE(debug_level, debug_message, __FILE__, __LINE__)

#define CPPTESK_DEBUG_PRINTF_FILE_LINE(debug_level, debug_message, ...) \
    __CPPTESK_DEBUG_PRINTF_FILE_LINE(debug_level, debug_message, __FILE__, __LINE__, __VA_ARGS__)

#ifdef HW_SHORT_NAMES
    #define DEBUG_PRINT(debug_level, debug_message)                 CPPTESK_DEBUG_PRINT(debug_level, debug_message)
    #define DEBUG_PRINTF(debug_level, debug_message, ...)           CPPTESK_DEBUG_PRINTF(debug_level, debug_message, __VA_ARGS__)
    #define DEBUG_PRINT_FILE_LINE(debug_level, debug_message)       CPPTESK_DEBUG_PRINT_FILE_LINE(debug_level, debug_message)
    #define DEBUG_PRINTF_FILE_LINE(debug_level, debug_message, ...) CPPTESK_DEBUG_PRINTF(debug_level, debug_message, __VA_ARGS__)
#endif // HW_SHORT_NAMES

namespace cpptesk {
namespace hw {

enum debug_level_t
{
    /// Print all messages.
    DEBUG_TOOL     =  5,
    DEBUG_INTERNAL =  5,

    ////////////////////////////////////////////////////////////////////////////
    // Internal Debug Levels
    ////////////////////////////////////////////////////////////////////////////
    DEBUG_ALL      =  4,
    ////////////////////////////////////////////////////////////////////////////
    // External Debug Levels
    ////////////////////////////////////////////////////////////////////////////

    /// Print all events including raw-level ones.
    DEBUG_MORE     =  4,
    DEBUG_DETAILED =  4,
    /// Print basic events only.
    DEBUG_INFO     =  3,
    DEBUG_BASIC    =  3,
    /// Print messages, warnings and errors.
    DEBUG_USER     =  2,
    DEBUG_MESSAGE  =  2,
    /// Print warnings and errors.
    DEBUG_WARN     =  1,
    DEBUG_WARNING  =  1,
    /// Print errors.
    DEBUG_FAIL     =  0,
    DEBUG_ERROR    =  0,
    /// Do not print debug messages.
    DEBUG_NONE     = -1,
};

class Debug
{
public:
    static void level(debug_level_t debug_level);
    static void color(bool debug_color);
    static void print(debug_level_t debug_level, const std::string &debug_message);
private:
    static const char *_STRING[DEBUG_TOOL + 1];
    static const char *_STYLE [DEBUG_TOOL + 1];
    static const char *_NORMAL_STYLE;
private:
    Debug() {}
   ~Debug() {}
private:
    static debug_level_t _debug_level;
    static bool _debug_color;
};

inline void debug(debug_level_t debug_level = DEBUG_ALL, bool debug_color = true)
    { Debug::level(debug_level); Debug::color(debug_color); }

}} // namespace cpptesk::hw

